# -*- coding:utf-8 -*-
from typing import Any, Optional, Union

from .base import BaseRecorder
from .style.cell_style import CellStyle
from .setter import RecorderSetter


class Recorder(BaseRecorder):
    _set: RecorderSetter = ...
    _col_height: Optional[float] = ...
    _follow_styles: bool = ...
    _style: Optional[CellStyle] = ...
    _quote_char: str = ...
    _delimiter: str = ...
    _data: Union[list, dict] = ...
    data: Union[list, dict] = ...
    _head: Optional[list, dict] = ...
    _fit_head: bool = ...

    @property
    def set(self) -> RecorderSetter: ...

    @property
    def delimiter(self) -> str: ...

    @property
    def quote_char(self) -> str: ...

    def add_data(self, data: Any, table: Union[str, bool] = None) -> None: ...

    def _record(self) -> None: ...

    def _to_xlsx(self) -> None: ...

    def _to_csv(self) -> None: ...

    def _to_txt(self) -> None: ...

    def _to_json(self) -> None: ...



# -*- coding:utf-8 -*-
from csv import reader as csv_reader, writer as csv_writer
from pathlib import Path
from time import sleep

from openpyxl import load_workbook, Workbook
from openpyxl.utils import get_column_letter

from .base import BaseRecorder
from .setter import FillerSetter
from .style.cell_style import CellStyle, NoneStyle
from .tools import parse_coord, get_usable_coord, process_content, data_to_list_or_dict


class Filler(BaseRecorder):
    def __init__(self, path=None, cache_size=None, key_cols=True, begin_row=2,
                 sign_col=True, data_col=None, sign=None, deny_sign=False):
        """用于处理表格文件的工具
        :param path: 保存的文件路径
        :param cache_size: 每接收多少条记录写入文件，传入0表示不自动保存
        :param key_cols: 作为关键字的列，可以是多列，可以是列编号或序号，从1开始，True表示获取整行
        :param begin_row: 数据开始的行，默认表头一行
        :param sign_col: 用于判断是否已填数据的列，可以是列编号或序号，从1开始，True表示获取所有行，不进行判断
        :param data_col: 要填入数据的第一列，从1开始
        :param sign: 按这个值筛选需要的行纳入keys
        :param deny_sign: 是否反向匹配sign，即筛选sign_col列值不是sign的行
        """
        super().__init__(None, cache_size)
        self._delimiter = ','  # csv文件分隔符
        self._quote_char = '"'  # csv文件引用符
        self._key_cols = None
        self._begin_row = None
        self._sign_col = None
        self._data_col = None
        self._sign = None
        self._deny_sign = False
        self.row_num_title = 'row'
        if not data_col:
            data_col = sign_col if sign_col else 1
        self.set.path(path, key_cols, begin_row, sign_col, data_col, sign, deny_sign)
        s = CellStyle()
        s.font.set_color("0000FF")
        s.font.set_underline('single')
        self._link_style = s

    @property
    def sign(self):
        """按这个值筛选需要的行纳入keys"""
        return self._sign

    @property
    def deny_sign(self):
        """返回是否反向匹配sign"""
        return self._deny_sign

    @property
    def key_cols(self):
        """返回作为关键字的列或列的集合"""
        return self._key_cols

    @property
    def sign_col(self):
        """返回用于判断是否已填数据的列"""
        return self._sign_col

    @property
    def data_col(self):
        """返回用于填充数据的列"""
        return self._data_col

    @property
    def begin_row(self):
        """返回数据开始的行号，用于获取keys，从1开始"""
        return self._begin_row

    @property
    def keys(self):
        """返回一个列表，由未执行的行数据组成。每行的格式为第一位为行号，其余为 key 列的值。
        eg.[3, '张三', 20]
        """
        if not self.path or not Path(self.path).exists():
            raise FileNotFoundError('未指定文件或文件不存在。')

        if self.type == 'csv':
            return get_csv_keys(self, False)
        elif self.type == 'xlsx':
            return get_xlsx_keys(self, False)

    @property
    def dict_keys(self):
        """返回一个列表，由未执行的行数据组成。每行的格式为dict，'row' 值为行号，其余值为第一行数据。
        如第一行数据为空，则用列号为值。如果begin_row为1，用列名作为值。
        eg.{'row': 2, 'name': '张三', 'C': '男'}
        """
        if not self.path or not Path(self.path).exists():
            raise FileNotFoundError('未指定文件或文件不存在。')

        if self.type == 'csv':
            return get_csv_keys(self, True)
        elif self.type == 'xlsx':
            return get_xlsx_keys(self, True)

    @property
    def set(self):
        """返回用于设置属性的对象"""
        if self._setter is None:
            self._setter = FillerSetter(self)
        return self._setter

    @property
    def delimiter(self):
        """返回csv文件分隔符"""
        return self._delimiter

    @property
    def quote_char(self):
        """返回csv文件引用符"""
        return self._quote_char

    def add_data(self, data, coord='newline', table=None):
        """添加数据，每次添加一行数据，可指定坐标、列号或行号
        coord只输入数字（行号）时，列号为self.data_col值，如 3；
        输入列号，或没有行号的坐标时，表示新增一行，列号为此时指定的，如'c'、',3'、(None, 3)、'None,3'；
        输入 'newline' 时，表示新增一行，列号为self.data_col值；
        输入行列坐标时，填写到该坐标，如'a3'、'3,1'、(3,1)、[3,1]；
        输入的行号列号可以是负数，代表从下往上数，-1是倒数第一行，如'a-3'、(-3, -3)
        :param data: 要添加的内容，任意格式
        :param coord: 要添加数据的坐标，可输入行号、列号或行列坐标，如'a3'、7、(3, 1)、[3, 1]、'c'
        :param table: 要写入的数据表，仅支持xlsx格式。为None表示用set.table()方法设置的值，为bool表示活动的表格
        :return: None
        """
        while self._pause_add:  # 等待其它线程写入结束
            sleep(.1)

        if not isinstance(data, (list, tuple)):
            data = (data,)

        if coord not in ('set_link', 'cover_style', 'replace_style', 'set_img', 'set_width', 'set_height'):
            coord = parse_coord(coord, self.data_col)
            if not data:
                data = ([],)
                self._data_count += 1
            # 一维数组
            elif isinstance(data, dict) or (
                    isinstance(data, (list, tuple)) and not isinstance(data[0], (list, tuple, dict))):
                data = (data_to_list_or_dict(self, data),)
                self._data_count += 1
            else:  # 二维数组
                data = [data_to_list_or_dict(self, d) for d in data]
                self._data_count += len(data)

        if self._type != 'xlsx':
            self._data.append((coord, data))

        else:
            if table is None:
                table = self._table
            elif isinstance(table, bool):
                table = None

            self._data.setdefault(table, []).append((coord, data))

        if 0 < self.cache_size <= self._data_count:
            self.record()

    def set_link(self, coord, link, content=None):
        """为单元格设置超链接
        :param coord: 单元格坐标
        :param link: 超链接，为None时删除链接
        :param content: 单元格内容
        :return: None
        """
        self.add_data((coord, link, content), 'set_link')

    def set_style(self, coord, style, replace=True):
        """为单元格设置样式
        :param coord: 单元格坐标，输入数字可设置整行，输入列名字符串可设置整列，输入'A1:C5'格式可设置指定范围
        :param style: CellStyle对象，为None则清除单元格样式
        :param replace: 是否直接替换已有样式，运行效率较高，但不能单独修改某个属性
        :return: None
        """
        s = 'replace_style' if replace else 'cover_style'
        self.add_data((coord, style), s)

    def set_img(self, coord, img_path, width=None, height=None):
        """
        :param coord: 单元格坐标，输入数字可设置整行，输入列名字符串可设置整列，输入'A1:C5'格式可设置指定范围
        :param img_path: 图片路径
        :param width: 图片宽
        :param height: 图片高
        :return: None
        """
        if isinstance(img_path, Path):
            img_path = str(img_path)
        self.add_data((coord, img_path, width, height), 'set_img')

    def set_row_height(self, row, height):
        """设置行高
        :param row: 行号
        :param height: 行高
        :return: None
        """
        self.add_data((row, height), 'set_height')

    def set_col_width(self, col, width):
        """设置列宽
        :param col: 列号，数字或字母
        :param width: 列宽
        :return: None
        """
        if isinstance(col, int):
            col = get_column_letter(col)
        self.add_data((col, width), 'set_width')

    def _record(self):
        """记录数据"""
        if self.type == 'xlsx':
            self._to_xlsx()
        elif self.type == 'csv':
            self._to_csv()

    def _to_xlsx(self):
        """填写数据到xlsx文件"""
        if Path(self.path).exists():
            wb = load_workbook(self.path)
            new = False
        else:
            wb = Workbook()
            new = True

        tables = [i.title for i in wb.worksheets]
        for table, table_data in self._data.items():
            if table is None:
                ws = wb.active

            else:
                if table in tables:
                    ws = wb[table]
                elif new is True:
                    new = False
                    ws = wb.active
                    ws.title = table
                    tables.append(table)
                else:
                    ws = wb.create_sheet(title=table)
                    tables.append(table)

            max_col = ws.max_column
            empty = ws.max_column == ws.max_row == 1 and ws[1][0].value is None

            for data in table_data:
                if empty:  # 如果是空文件，最大行数设为0，避免直接添加时出现空行
                    max_row = 0
                    empty = False
                else:
                    max_row = ws.max_row

                if data[0] == 'set_link':
                    coord = parse_coord(data[1][0], self.data_col)
                    row, col = get_usable_coord(coord, max_row, max_col)
                    cell = ws.cell(row, col)
                    has_link = True if cell.hyperlink else Filler
                    cell.hyperlink = None if data[1][1] is None else process_content(data[1][1], True)
                    if data[1][2] is not None:
                        cell.value = process_content(data[1][2], True)
                    if data[1][1]:
                        if self._link_style:
                            self._link_style.to_cell(cell, replace=False)
                    elif has_link:
                        NoneStyle().to_cell(cell, replace=False)
                    continue

                elif data[0] in ('replace_style', 'cover_style'):
                    mode = data[0] == 'replace_style'
                    coord = data[1][0]
                    style = NoneStyle() if data[1][1] is None else data[1][1]
                    if isinstance(coord, int) or (isinstance(coord, str) and coord.isdigit()):
                        for c in ws[coord]:
                            style.to_cell(c, replace=mode)
                        continue

                    elif isinstance(coord, str) and ':' in coord:
                        for c in ws[coord]:
                            for cc in c:
                                style.to_cell(cc, replace=mode)
                        continue

                    coord = parse_coord(coord, self.data_col)
                    row, col = get_usable_coord(coord, max_row, max_col)
                    style.to_cell(ws.cell(row, col), replace=mode)
                    continue

                elif data[0] == 'set_img':
                    coord, img_path, width, height = data[1]
                    coord = parse_coord(coord, self.data_col)
                    row, col = get_usable_coord(coord, max_row, max_col)

                    from openpyxl.drawing.image import Image
                    from openpyxl.utils import get_column_letter
                    img = Image(img_path)
                    if width and height:
                        img.width = width
                        img.height = height
                    elif width:
                        img.height = int(img.height * (width / img.width))
                        img.width = width
                    elif height:
                        img.width = int(img.width * (height / img.height))
                        img.height = height
                    col = get_column_letter(col)
                    ws.add_image(img, f'{col}{row}')
                    continue

                elif data[0] == 'set_width':
                    col, width = data[1]
                    ws.column_dimensions[col].width = width
                    continue

                elif data[0] == 'set_height':
                    row, height = data[1]
                    ws.row_dimensions[row].height = height
                    continue

                row, col = get_usable_coord(data[0], max_row, max_col)
                now_data = (data[1],) if not isinstance(data[1][0], (list, tuple, dict)) else data[1]

                for r, i in enumerate(now_data, row):
                    if isinstance(i, dict):
                        i = i.values()
                    for key, j in enumerate(i):
                        ws.cell(r, col + key).value = process_content(j, True)

        wb.save(self.path)
        wb.close()

    def _to_csv(self):
        """填写数据到xlsx文件"""
        if not Path(self.path).exists():
            with open(self.path, 'w', encoding=self.encoding):
                pass

        with open(self.path, 'r', encoding=self.encoding) as f:
            reader = csv_reader(f, delimiter=self.delimiter, quotechar=self.quote_char)
            lines = list(reader)
            lines_count = len(lines)

            for i in self._data:
                if i[0] == 'set_link':
                    coord = parse_coord(i[1][0], self.data_col)
                    now_data = (f'=HYPERLINK("{i[1][1]}","{i[1][2] or i[1][1]}")',)

                elif i[0] in ('cover_style', 'replace_style', 'set_img', 'set_width', 'set_height'):
                    continue

                else:
                    coord = i[0]
                    now_data = i[1]

                row, col = get_usable_coord(coord, lines_count, len(lines[0]) if lines_count else 1)
                now_data = (now_data,) if not isinstance(now_data[0], (list, tuple, dict)) else now_data

                for r, data in enumerate(now_data, row):
                    if isinstance(data, dict):
                        data = list(data.values())

                    for _ in range(r - lines_count):  # 若行数不够，填充行数
                        lines.append([])
                        lines_count += 1

                    row_num = r - 1

                    # 若列数不够，填充空列
                    lines[row_num].extend([None] * (col - len(lines[row_num]) + len(data) - 1))

                    for k, j in enumerate(data):  # 填充数据
                        lines[row_num][col + k - 1] = process_content(j)

            writer = csv_writer(open(self.path, 'w', encoding=self.encoding, newline=''), delimiter=self.delimiter,
                                quotechar=self.quote_char)
            writer.writerows(lines)


def get_xlsx_keys(filler, as_dict):
    """返回key列内容，第一位为行号，其余为key列的值
    如果as_dict为True，返回dict格式，value为第一行值，值为空或begin_row为1时用列号，'row' 值为行号
    eg.[3, '名称', 'id']
    :param filler: 记录器对象
    :param as_dict: 是否以dict格式返回数据
    :return: 关键字组成的列表或字典
    """
    wb = load_workbook(filler.path, data_only=True, read_only=True)
    if filler.table and filler.table not in [i.title for i in wb.worksheets]:
        raise RuntimeError(f'xlsx文件未包含此工作表：{filler.table}')
    ws = wb[filler.table] if filler.table else wb.active

    if ws.max_column is None:  # 遇到过read_only时无法获取列数的文件
        wb.close()
        wb = load_workbook(filler.path, data_only=True)
        ws = wb[filler.table] if filler.table else wb.active

    rows = ws.rows
    loop = 0
    if as_dict:
        title = [filler.row_num_title]
        if filler.begin_row == 1:
            t = [get_column_letter(x) for x in range(1, ws.max_column + 1)
                 if filler.key_cols is True or x in filler.key_cols]
        else:
            try:
                t = next(rows)
                t = [x.value if x.value else get_column_letter(k) for k, x in enumerate(t, 1)
                     if filler.key_cols is True or k in filler.key_cols]
                loop = 1

            except StopIteration:
                return []

        if len(t) != len(set(t)):
            raise RuntimeError('表头出现内容重复。')
        if filler.row_num_title in t:
            raise RuntimeError(f'表头"{filler.row_num_title}"列与行号列重复，请用set.row_num_title()方法设置不同的表头。')
        title.extend(t)

    try:
        for _ in range(filler.begin_row - loop - 1):
            next(rows)
    except StopIteration:
        return []

    # ---------------------------------------------------------

    if filler.sign_col is True or filler.sign_col > ws.max_column:  # 获取所有行
        if filler.key_cols is True:  # 获取整行
            res = [[ind] + [i.value for i in row]
                   for ind, row in enumerate(rows, filler.begin_row)]
        else:  # 只获取对应的列
            res = [[ind] + [row[i - 1].value for i in filler.key_cols]
                   for ind, row in enumerate(rows, filler.begin_row)]

    else:  # 获取符合条件的行
        if filler.key_cols is True:  # 获取整行
            if filler.deny_sign:
                res = [[ind] + [i.value for i in row]
                       for ind, row in enumerate(rows, filler.begin_row)
                       if row[filler.sign_col - 1].value != filler.sign]
            else:
                res = [[ind] + [i.value for i in row]
                       for ind, row in enumerate(rows, filler.begin_row)
                       if row[filler.sign_col - 1].value == filler.sign]

        else:  # 只获取对应的列
            if filler.deny_sign:
                res = [[ind] + [row[i - 1].value for i in filler.key_cols]
                       for ind, row in enumerate(rows, filler.begin_row)
                       if row[filler.sign_col - 1].value != filler.sign]
            else:
                res = [[ind] + [row[i - 1].value for i in filler.key_cols]
                       for ind, row in enumerate(rows, filler.begin_row)
                       if row[filler.sign_col - 1].value == filler.sign]

    wb.close()

    if as_dict:
        res = [dict(zip(title, v)) for v in res]
    return res


def get_csv_keys(filler, as_dict):
    """返回key列内容，第一位为行号，其余为key列的值，
    如果as_dict为True，返回dict格式，value为第一行值，值为空或begin_row为1时用列号，'row'值为行号
    eg.[3, '名称', 'id']
    :param filler: 记录器对象
    :param as_dict: 是否以dict格式返回数据
    :return: 关键字组成的列表或字典
    """
    begin_row = filler.begin_row
    sign_col = filler.sign_col
    sign = '' if filler.sign is None else str(filler.sign)
    begin_row -= 1
    res = []

    with open(filler.path, 'r', encoding=filler.encoding) as f:
        reader = csv_reader(f, delimiter=filler.delimiter, quotechar=filler.quote_char)
        lines = list(reader)
        if not lines:
            return res

        if as_dict:
            title = [filler.row_num_title]
            if filler.begin_row == 1:
                t = [get_column_letter(x) for x in range(1, len(lines[0]) + 1)
                     if filler.key_cols is True or x in filler.key_cols]
            else:
                t = [x if x else get_column_letter(k) for k, x in enumerate(lines[0], 1)
                     if filler.key_cols is True or k in filler.key_cols]

            if len(t) != len(set(t)):
                raise RuntimeError('表头内容重复。')
            if filler.row_num_title in t:
                raise RuntimeError(f'表头"{filler.row_num_title}"列与行号列重复，请用set.row_num_title()方法设置不同的表头。')
            title.extend(t)

        if sign_col is not True:  # 获取符合条件的行
            sign_col -= 1
            for ind, line in enumerate(lines[begin_row:], begin_row + 1):
                row_sign = '' if sign_col > len(line) - 1 else line[sign_col]
                if row_sign != sign if filler.deny_sign else row_sign == sign:
                    if filler.key_cols is True:  # 获取整行
                        res.append([ind] + line)
                    else:  # 只获取对应的列
                        res.append([ind] + [line[i - 1] for i in filler.key_cols])

        else:  # 获取所有行
            for ind, line in enumerate(lines[begin_row:], begin_row + 1):
                if filler.key_cols is True:  # 获取整行
                    res.append([ind] + line)
                else:  # 只获取对应的列
                    res.append([ind] + [line[i - 1] for i in filler.key_cols])

    if as_dict:
        res = [dict(zip(title, v)) for v in res]
    return res



# -*- coding:utf-8 -*-
from pathlib import Path

from openpyxl.reader.excel import load_workbook
from openpyxl.utils import column_index_from_string
from openpyxl.workbook import Workbook

from .tools import process_content, ok_list, make_valid_name


class OriginalSetter(object):
    def __init__(self, recorder):
        self._recorder = recorder

    def cache_size(self, size):
        """设置缓存大小
        :param size: 缓存大小
        :return: None
        """
        if not isinstance(size, int) or size < 0:
            raise TypeError('cache_size值只能是int，且必须>=0')
        self._recorder._cache = size

    def path(self, path):
        """设置文件路径
        :param path: 文件路径
        :return: None
        """
        if self._recorder._path:
            self._recorder.record()

        p = Path(path)
        self._recorder._path = str(p.parent / make_valid_name(p.name))
        self._recorder._data = []

    def show_msg(self, on_off):
        """设置是否显示运行信息
        :param on_off: bool表示开关
        :return: None
        """
        self._recorder.show_msg = on_off


class BaseSetter(OriginalSetter):
    def table(self, name):
        """设置默认表名
        :param name: 表名
        :return: None
        """
        self._recorder._table = name

    def before(self, before):
        """设置在数据前面补充的列
        :param before: 列表、元组或字符串，为字符串时则补充一列
        :return: None
        """
        if before is None:
            self._recorder._before = None
        elif isinstance(before, (list, dict)):
            self._recorder._before = before
        elif isinstance(before, tuple):
            self._recorder._before = list(before)
        else:
            self._recorder._before = [before]

    def after(self, after):
        """设置在数据后面补充的列
        :param after: 列表、元组或字符串，为字符串时则补充一列
        :return: None
        """
        if after is None:
            self._recorder._after = None
        elif isinstance(after, (list, dict)):
            self._recorder._after = after
        elif isinstance(after, tuple):
            self._recorder._after = list(after)
        else:
            self._recorder._after = [after]

    def encoding(self, encoding):
        """设置编码
        :param encoding: 编码格式
        :return: None
        """
        self._recorder._encoding = encoding


class SheetLikeSetter(BaseSetter):
    def head(self, head, table=None):
        """设置表头。只有 csv 和 xlsx 格式支持设置表头
        :param head: 表头，列表或元组
        :param table: 表名，只xlsx格式文件有效
        :return: None
        """
        if not self._recorder.path:
            raise FileNotFoundError('未指定文件。')
        if not isinstance(head, (list, tuple)):
            raise TypeError('head参数只能是list或tuple格式。')

        if self._recorder.type == 'xlsx':
            table = table or self._recorder.table
            set_xlsx_head(self._recorder.path, head, table)

        elif self._recorder.type == 'csv':
            set_csv_head(self._recorder.path, head, self._recorder.encoding, self._recorder.delimiter,
                         self._recorder.quote_char)

        else:
            raise TypeError('只能对xlsx和csv文件设置表头。')

    def delimiter(self, delimiter):
        """设置csv文件分隔符
        :param delimiter: 分隔符
        :return: None
        """
        self._recorder._delimiter = delimiter

    def quote_char(self, quote_char):
        """设置csv文件引用符
        :param quote_char: 引用符
        :return: None
        """
        self._recorder._quote_char = quote_char

    def path(self, path, file_type=None):
        """设置文件路径
        :param path: 文件路径
        :param file_type: 要设置的文件类型，为空则从文件名中获取
        :return: None
        """
        super().path(path)

        if not file_type:
            suffix = Path(path).suffix.lower()
            if suffix:
                file_type = suffix[1:]
            elif not self._recorder.type:
                file_type = 'csv'

        if file_type:
            self.file_type(file_type)

        self._recorder._data = {} if self._recorder._type == 'xlsx' else []

    def file_type(self, file_type):
        """指定文件类型，无视文件后缀名"""
        if 'any' not in self._recorder.SUPPORTS and file_type not in self._recorder.SUPPORTS:
            raise TypeError(f'只支持{"、".join(self._recorder.SUPPORTS)}格式文件。')
        self._recorder._type = file_type

    def table(self, name):
        """设置默认表名
        :param name: 表名
        :return: None
        """
        if isinstance(name, bool):
            name = None
        self._recorder._table = name


class FillerSetter(SheetLikeSetter):
    def sign(self, value):
        """设置sign值
        :param value: 筛选条件
        :return: None
        """
        self._recorder._sign = value

    def deny_sign(self, on_off=True):
        """设置是否反向匹配sign
        :param on_off: bool表示开或关
        :return: None
        """
        self._recorder._deny_sign = on_off

    def key_cols(self, cols):
        """设置作为关键字的列，可以是多列
        :param cols: 列号或列名，或它们组成的list或tuple
        :return: None
        """
        if cols is True:
            self._recorder._key_cols = True
        elif isinstance(cols, int) and cols > 0:
            self._recorder._key_cols = [cols]
        elif isinstance(cols, str):
            self._recorder._key_cols = [int(cols)] if cols.isdigit() else [column_index_from_string(cols)]
        elif isinstance(cols, (list, tuple)):
            self._recorder._key_cols = [i if isinstance(i, int) and i > 0 else
                                        int(i) if i.isdigit() else column_index_from_string(i) for i in cols]
        else:
            raise TypeError('col值只能是int或str，且必须大于0。')

    def sign_col(self, col):
        """设置用于判断是否已填数据的列
        :param col: 列号或列名
        :return: None
        """
        if col is True or (isinstance(col, int) and col > 0):
            self._recorder._sign_col = col
        elif isinstance(col, str):
            self._recorder._sign_col = int(col) if col.isdigit() else column_index_from_string(col)
        else:
            raise TypeError('col值只能是True、int或str，且必须大于0。')

    def data_col(self, col):
        """设置用于填充数据的列
        :param col: 列号或列名
        :return: None
        """
        if isinstance(col, int) and col > 0:
            self._recorder._data_col = col
        elif isinstance(col, str):
            self._recorder._data_col = column_index_from_string(col)
        else:
            raise TypeError('col值只能是int或str，且必须大于0。')

    def begin_row(self, row):
        """设置数据开始的行
        :param row: 行号
        :return: None
        """
        if not isinstance(row, int) or row < 1:
            raise TypeError('row值只能是int，且必须大于0')
        self._recorder._begin_row = row

    def path(self, path=None, key_cols=None, begin_row=None, sign_col=None,
             data_col=None, sign=None, deny_sign=None):
        """设置文件路径
        :param path: 保存的文件路径
        :param key_cols: 作为关键字的列，可以是多列
        :param begin_row: 数据开始的行，默认表头一行
        :param sign_col: 用于判断是否已填数据的列
        :param data_col: 要填入数据的第一列
        :param sign: 按这个值判断是否已填数据
        :param deny_sign: 是否反向匹配sign，即筛选指不是sign的行
        """
        if path:
            super().path(path)
        self._recorder.set.key_cols(key_cols or self._recorder.key_cols)
        self._recorder.set.begin_row(begin_row or self._recorder.begin_row)
        self._recorder.set.sign_col(sign_col or self._recorder.sign_col)
        self._recorder.set.sign(sign or self._recorder.sign)
        self._recorder.set.data_col(data_col or self._recorder.data_col)
        self._recorder.set.deny_sign(deny_sign if deny_sign is not None else self._recorder.deny_sign)

    def link_style(self, style):
        """设置单元格的链接样式
        :param style: CellStyle对象
        :return: None
        """
        self._recorder._link_style = style

    def row_num_title(self, title):
        """设置dict_keys返回的数据中，行号的键名
        :param title: 行号标题
        :return: None
        """
        self._recorder.row_num_title = title


class RecorderSetter(SheetLikeSetter):
    def follow_styles(self, on_off=True):
        """设置是否跟随最后一行的style，只有xlsx格式有效
        :param on_off: True或False
        :return: None
        """
        self._recorder._follow_styles = on_off

    def col_height(self, height):
        """设置行高，只有xlsx格式有效
        :param height: 行高
        :return: None
        """
        self._recorder._col_height = height

    def style(self, style):
        """设置样式，只有xlsx格式有效
        :param style: CellStyle对象
        :return: None
        """
        self._recorder._style = style

    def path(self, path, file_type=None):
        """设置文件路径
        :param path: 文件路径
        :param file_type: 要设置的文件类型，为空则从文件名中获取
        :return: None
        """
        super().path(path=path, file_type=file_type)
        self._recorder._row_styles = None
        self._recorder._head = {} if self._recorder.type == 'xlsx' else None

    def head(self, head, table=None):
        """设置表头。只有 csv 和 xlsx 格式支持设置表头
        :param head: 表头，列表或元组
        :param table: 表名，只xlsx格式文件有效
        :return: None
        """
        super().head(head, table)
        if self._recorder.type == 'xlsx':
            if not self._recorder._head or not isinstance(self._recorder._head, dict):
                self._recorder._head = {table: head}
            else:
                self._recorder._head[table] = head

        elif self._recorder.type == 'csv':
            self._recorder._head = head

    def fit_head(self, on_off=True):
        """设置是否自动匹配表头
        :param on_off: bool表示开关
        :return: None
        """
        if self._recorder.type not in ('csv', 'xlsx'):
            raise TypeError('只有csv或xlsx格式可设置fit_head。')
        self._recorder._fit_head = on_off


class DBSetter(BaseSetter):
    def path(self, path, table=None):
        """重写父类方法
        :param path: 文件路径
        :param table: 数据表名称
        :return: None
        """
        with self._recorder._lock:
            super().path(path)
            if self._recorder._conn is not None:
                self._recorder._close_connection()
            self._recorder._connect()

            if table:
                self._recorder._table = table

            else:
                r = self._recorder.run_sql("select name from sqlite_master where type='table'")
                self._recorder._table = r[0] if r else None

            self._recorder._data = {}
            self._recorder._close_connection()


def set_csv_head(file_path, head, encoding='utf-8', delimiter=',', quote_char='"'):
    """设置csv文件的表头
    :param file_path: 文件路径
    :param head: 表头列表或元组
    :param encoding: 编码
    :param delimiter: 分隔符
    :param quote_char: 引用符
    :return: None
    """
    from csv import writer
    if Path(file_path).exists():
        with open(file_path, 'r', newline='', encoding=encoding) as f:
            content = "".join(f.readlines()[1:])

        with open(file_path, 'w', newline='', encoding=encoding) as f:
            csv_write = writer(f, delimiter=delimiter, quotechar=quote_char)
            csv_write.writerow(ok_list(head))

        with open(file_path, 'a+', newline='', encoding=encoding) as f:
            f.write(f'{content}')

    else:
        Path(file_path).parent.mkdir(parents=True, exist_ok=True)
        with open(file_path, 'w', newline='', encoding=encoding) as f:
            csv_write = writer(f, delimiter=delimiter, quotechar=quote_char)
            csv_write.writerow(ok_list(head))


def set_xlsx_head(file_path, head, table):
    """设置xlsx文件的表头
    :param file_path: 文件路径
    :param head: 表头列表或元组
    :param table: 工作表名称
    :return: None
    """
    if Path(file_path).exists():
        wb = load_workbook(file_path)
        if table:
            ws = wb[table] if table in [i.title for i in wb.worksheets] else wb.create_sheet(title=table)
        else:
            ws = wb.active

    else:
        Path(file_path).parent.mkdir(parents=True, exist_ok=True)
        wb = Workbook()
        ws = wb.active
        if table:
            ws.title = table

    if len(ws[1]) > len(head):
        head = list(head)
        head.extend([None] * (len(ws[1]) - len(head)))

    for key, i in enumerate(head, 1):
        ws.cell(1, key).value = process_content(i, True)

    wb.save(file_path)
    wb.close()



# -*- coding:utf-8 -*-
from pathlib import Path
from time import sleep
from typing import Union

from openpyxl import Workbook, load_workbook

from .base import BaseRecorder
from .style.cell_style import CellStyleCopier
from .setter import RecorderSetter
from .tools import ok_list, data_to_list_or_dict, process_content


class Recorder(BaseRecorder):
    SUPPORTS = ('csv', 'xlsx', 'json', 'txt')

    def __init__(self, path=None, cache_size=None):
        """用于缓存并记录数据，可在达到一定数量时自动记录，以降低文件读写次数，减少开销
        :param path: 保存的文件路径
        :param cache_size: 每接收多少条记录写入文件，0为不自动写入
        """
        super().__init__(path=path, cache_size=cache_size)
        self._delimiter = ','  # csv文件分隔符
        self._quote_char = '"'  # csv文件引用符
        self._follow_styles = False
        self._col_height = None
        self._style = None
        self._fit_head = False

    @property
    def set(self):
        """返回用于设置属性的对象"""
        if self._setter is None:
            self._setter = RecorderSetter(self)
        return self._setter

    @property
    def delimiter(self):
        """返回csv文件分隔符"""
        return self._delimiter

    @property
    def quote_char(self):
        """返回csv文件引用符"""
        return self._quote_char

    def add_data(self, data, table=None):
        """添加数据，可一次添加多条数据
        :param data: 插入的数据，任意格式
        :param table: 要写入的数据表，仅支持xlsx格式。为None表示用set.table()方法设置的值，为bool表示活动的表格
        :return: None
        """
        while self._pause_add:  # 等待其它线程写入结束
            sleep(.1)

        if not isinstance(data, (list, tuple, dict)):
            data = (data,)

        if not data:
            data = ([],)
            self._data_count += 1

        # 一维数组
        elif isinstance(data, dict) or (
                isinstance(data, (list, tuple)) and not isinstance(data[0], (list, tuple, dict))):
            data = [data_to_list_or_dict(self, data)]
            self._data_count += 1

        else:  # 二维数组
            data = [data_to_list_or_dict(self, d) for d in data]
            self._data_count += len(data)

        if self._type != 'xlsx':
            self._data.extend(data)

        else:
            if table is None:
                table = self._table
            elif isinstance(table, bool):
                table = None

            self._data.setdefault(table, []).extend(data)

        if 0 < self.cache_size <= self._data_count:
            self.record()

    def _record(self):
        """记录数据"""
        if self.type == 'csv':
            self._to_csv()
        elif self.type == 'xlsx':
            self._to_xlsx()
        elif self.type == 'json':
            self._to_json()
        elif self.type == 'txt':
            self._to_txt()

    def _to_xlsx(self):
        """记录数据到xlsx文件"""
        if Path(self.path).exists():
            new_file = False
            wb = load_workbook(self.path)

        else:
            new_file = True
            if self._col_height or self._follow_styles or self._style or len(self._data) > 1:
                wb = Workbook(write_only=False)
            else:
                wb = Workbook(write_only=True)
                wb.create_sheet('Sheet1')

        tables = [i.title for i in wb.worksheets]
        for table, data in self._data.items():
            _row_styles = None
            _col_height = None
            new_sheet = False

            if table is None:
                ws = wb.active

            elif table in tables:
                ws = wb[table]

            elif new_file:
                ws = wb.active
                tables.remove(ws.title)
                ws.title = table
                tables.append(table)
                new_sheet = True

            else:
                ws = wb.create_sheet(title=table)
                tables.append(table)
                new_sheet = True

            if new_file or new_sheet:
                new_file = False
                title = _get_title(data[0], self._before, self._after)
                if title is not None:
                    ws.append(ok_list(title, True))

            elif self._follow_styles:
                row_num = ws.max_row
                _row_styles = [CellStyleCopier(i) for i in ws[row_num]]
                _col_height = ws.row_dimensions[row_num].height
                title = [r.value for r in ws[1]]

            else:
                title = [r.value for r in ws[1]]

            if self._fit_head and not self._head.get(ws.title, None) and any(title):
                self._head[ws.title] = title

            # ====================================

            if self._fit_head and self._head.get(ws.title, None):
                for i in data:
                    if isinstance(i, dict):
                        i = [i.get(h, None) for h in self._head[ws.title]]

                    ws.append(ok_list(i, True))
                    _set_style(_col_height, _row_styles, ws, self)

            else:
                for i in data:
                    ws.append(ok_list(i, True))
                    _set_style(_col_height, _row_styles, ws, self)

        wb.save(self.path)
        wb.close()

    def _to_csv(self):
        """记录数据到csv文件"""
        if Path(self.path).exists():
            title = None
            if self._fit_head and not self._head:
                from csv import reader
                with open(self.path, 'r', newline='', encoding=self.encoding) as f:
                    u = reader(f, delimiter=self.delimiter, quotechar=self.quote_char)
                    try:
                        self._head = next(u)
                    except StopIteration:
                        pass

        else:
            title = _get_title(self._data[0], self._before, self._after)
            if self._fit_head:
                self._head = title

        with open(self.path, 'a+', newline='', encoding=self.encoding) as f:
            from csv import writer
            csv_write = writer(f, delimiter=self.delimiter, quotechar=self.quote_char)
            if title:
                csv_write.writerow(ok_list(title))

            if self._fit_head and self._head:
                for i in self._data:
                    d = data_to_list_or_dict(self, i)
                    if isinstance(d, dict):
                        d = [d.get(h, None) for h in self._head]

                    csv_write.writerow(ok_list(d))

            else:
                for i in self._data:
                    csv_write.writerow(ok_list(i))

    def _to_txt(self):
        """记录数据到txt文件"""
        with open(self.path, 'a+', encoding=self.encoding) as f:
            all_data = [' '.join(ok_list(i, as_str=True)) for i in self._data]
            f.write('\n'.join(all_data) + '\n')

    def _to_json(self):
        """记录数据到json文件"""
        from json import load, dump
        if Path(self.path).exists():
            with open(self.path, 'r', encoding=self.encoding) as f:
                json_data = load(f)

            for i in self._data:
                if isinstance(i, dict):
                    for d in i:
                        i[d] = process_content(i[d])
                    json_data.append(i)
                else:
                    json_data.append([process_content(d) for d in i])

        else:
            for i in self._data:
                if isinstance(i, dict):
                    for d in i:
                        i[d] = process_content(i[d])
                    json_data = i
                else:
                    json_data = [process_content(d) for d in i]

        with open(self.path, 'w', encoding=self.encoding) as f:
            dump(json_data, f)


def _set_style(_col_height, _row_styles, ws, recorder):
    if _col_height is not None:
        ws.row_dimensions[ws.max_row].height = recorder._col_height

    if _row_styles:
        groups = zip(ws[ws.max_row], _row_styles)
        for g in groups:
            g[1].to_cell(g[0])

    elif recorder._style:
        for c in ws[ws.max_row]:
            recorder._style.to_cell(c)


def _get_title(data: Union[list, dict],
               before: Union[list, dict, None] = None,
               after: Union[list, dict, None] = None) -> Union[list, None]:
    """获取表头列表
    :param data: 数据列表或字典
    :param before: 数据前的列
    :param after: 数据后的列
    :return: 表头列表
    """
    if isinstance(data, (tuple, list)):
        return None

    return_list = []
    for i in (before, data, after):
        if isinstance(i, dict):
            return_list.extend(list(i))
        elif i is None:
            pass
        elif isinstance(i, list):
            return_list.extend(['' for _ in range(len(i))])
        else:
            return_list.extend([''])

    return return_list



# -*- coding:utf-8 -*-
from abc import abstractmethod
from pathlib import Path
from threading import Lock
from typing import Union, Any, Optional

from .setter import OriginalSetter, BaseSetter


class OriginalRecorder(object):
    SUPPORTS: tuple = ...
    _cache: int = ...
    _path: Optional[str] = ...
    _type: Optional[str] = ...
    _data: Optional[list, dict] = ...
    _lock: Lock = ...
    _pause_add: bool = ...
    _pause_write: bool = ...
    show_msg: bool = ...
    _setter: Optional[OriginalSetter] = ...
    _data_count: int = ...

    def __init__(self,
                 path: Optional[str, Path] = None,
                 cache_size: int = None) -> None: ...

    def __del__(self) -> None: ...

    @property
    def set(self) -> OriginalSetter: ...

    @property
    def cache_size(self) -> int: ...

    @property
    def path(self) -> str: ...

    @property
    def type(self) -> str: ...

    @property
    def data(self) -> Union[dict, list]: ...

    def record(self, new_path: Optional[str, Path] = None) -> str: ...

    def clear(self) -> None: ...

    @abstractmethod
    def add_data(self, data): ...

    @abstractmethod
    def _record(self): ...


class BaseRecorder(OriginalRecorder):
    SUPPORTS: tuple = ...
    _encoding: str = ...
    _before: list = ...
    _after: list = ...
    _table: Optional[str] = ...
    _setter: BaseSetter = ...

    def __init__(self, path: Optional[str, Path] = None, cache_size: int = None) -> None: ...

    @property
    def set(self) -> BaseSetter: ...

    @property
    def before(self) -> Any: ...

    @property
    def after(self) -> Any: ...

    @property
    def table(self) -> Optional[str]: ...

    @property
    def encoding(self) -> str: ...

    @abstractmethod
    def _record(self): ...



# -*- coding:utf-8 -*-
from pathlib import Path
from sqlite3 import connect
from time import sleep

from .base import BaseRecorder
from .setter import DBSetter
from .tools import data_to_list_or_dict, ok_list


class DBRecorder(BaseRecorder):
    SUPPORTS = ('db',)

    def __init__(self, path=None, cache_size=None, table=None):
        """用于存储数据到sqlite的工具
        :param path: 保存的文件路径
        :param cache_size: 每接收多少条记录写入文件，0为不自动写入
        :param table: 默认表名
        """
        self._conn = None
        self._cur = None
        self._type = 'db'
        super().__init__(None, cache_size)
        if path:
            self.set.path(path, table)

    @property
    def set(self):
        """返回用于设置属性的对象"""
        if self._setter is None:
            self._setter = DBSetter(self)
        return self._setter

    def add_data(self, data, table=None):
        """添加数据
        :param data: 可以是一维或二维数据，dict格式可向对应列填写数据，其余格式按顺序从左到右填入各列
        :param table: 数据要插入的表名称
        :return: None
        """
        while self._pause_add:  # 等待其它线程写入结束
            sleep(.1)

        table = table or self.table
        if not isinstance(table, str):
            raise RuntimeError('未指定数据库表名。')

        if not isinstance(data, (list, tuple, dict)):
            data = (data,)

        if not data:
            self._data.setdefault(table, []).append(tuple())
            self._data_count += 1

        # 一维数组
        elif isinstance(data, dict) or (
                isinstance(data, (list, tuple)) and not isinstance(data[0], (list, tuple, dict))):
            self._data.setdefault(table, []).append(data_to_list_or_dict(self, data))
            self._data_count += 1

        else:  # 二维数组
            self._data.setdefault(table, []).extend([data_to_list_or_dict(self, d) for d in data])
            self._data_count += len(data)

        if 0 < self.cache_size <= self._data_count:
            self.record()

    def run_sql(self, sql, single=True, commit=False):
        """执行sql语句并返回结果
        :param sql: sql语句
        :param single: 是否只获取一个结果
        :param commit: 是否提交到数据库
        :return: 查找到的结果，没有结果时返回None
        """
        self._connect()
        self._cur.execute(sql)
        r = self._cur.fetchone() if single else self._cur.fetchall()
        if commit:
            self._conn.commit()
        self._close_connection()
        return r

    def _connect(self):
        """连接数据库"""
        path = Path(self.path).parent
        if not path.exists():
            path.mkdir(parents=True, exist_ok=True)
        self._conn = connect(self.path)
        self._cur = self._conn.cursor()

    def _close_connection(self):
        """关闭数据库 """
        if self._conn is not None:
            try:
                self._cur.close()
                self._conn.close()
            except:
                pass

    def _to_database(self, data_list, table, tables):
        """把数据批量写入指定数据表
        :param data_list: 要写入的数据组成的列表
        :param table: 要写入数据的数据表名称
        :param tables: 数据库中数据表和列信息
        :return: None
        """
        if isinstance(data_list[0], dict):  # 检查是否要新增列
            keys = data_list[0].keys()
            for key in keys:
                if key not in tables[table]:
                    sql = f'ALTER TABLE {table} ADD COLUMN {key}'
                    self._cur.execute(sql)
                    tables[table].append(key)

            question_masks = ','.join('?' * len(keys))
            keys_txt = ','.join(keys)
            values = [ok_list(i.values()) for i in data_list]
            sql = f'INSERT INTO {table} ({keys_txt}) values ({question_masks})'

        else:
            question_masks = ','.join('?' * len(tables[table]))
            values = data_list
            sql = f'INSERT INTO {table} values ({question_masks})'

        self._cur.executemany(sql, values)

    def _record(self):
        """保存数据到sqlite"""
        # 获取所有表名和列名
        self._connect()
        self._cur.execute("select name from sqlite_master where type='table'")
        tables = {}
        for table in self._cur.fetchall():
            self._cur.execute(f"PRAGMA table_info({table[0]})")
            tables[table[0]] = [i[1] for i in self._cur.fetchall()]

        for table, data in self._data.items():
            data_list = []
            if isinstance(data[0], dict):
                curr_keys = data[0].keys()
            else:
                curr_keys = len(data[0])

            for d in data:
                if isinstance(d, dict):
                    tmp_keys = d.keys()
                    if table not in tables:
                        keys = d.keys()
                        self._cur.execute(f"CREATE TABLE {table} ({','.join(keys)})")
                        tables[table] = tuple(keys)

                else:
                    if table not in tables:
                        self._close_connection()
                        raise TypeError('新建表格首次须接收数据需为dict格式。')
                    tmp_keys = len(d)
                    long = len(tables[table])
                    if long > tmp_keys:
                        d = ok_list(d)
                        d.extend([None] * (long - tmp_keys))
                    elif long < tmp_keys:
                        self._close_connection()
                        raise RuntimeError('数据个数大于列数（注意before和after属性）。')

                if tmp_keys != curr_keys:
                    self._to_database(data_list, table, tables)
                    curr_keys = tmp_keys
                    data_list = []

                data_list.append(d)

            if data_list:
                self._to_database(data_list, table, tables)

        self._conn.commit()
        self._close_connection()



# -*- coding:utf-8 -*-
from pathlib import Path
from time import sleep

from .base import OriginalRecorder


class ByteRecorder(OriginalRecorder):
    SUPPORTS = ('any',)
    __END = (0, 2)

    def __init__(self, path=None, cache_size=None):
        """用于记录字节数据的工具
        :param path: 保存的文件路径
        :param cache_size: 每接收多少条记录写入文件，0为不自动写入
        """
        super().__init__(path, cache_size)

    def add_data(self, data, seek=None):
        """添加一段二进制数据
        :param data: bytes类型数据
        :param seek: 在文件中的位置，None表示最后
        :return: None
        """
        while self._pause_add:  # 等待其它线程写入结束
            sleep(.1)

        if not isinstance(data, bytes):
            raise TypeError('只能接受bytes类型数据。')
        if seek is not None and not (isinstance(seek, int) and seek >= 0):
            raise ValueError('seek参数只能接受None或大于等于0的整数。')

        self._data.append((data, seek))
        self._data_count += 1

        if 0 < self.cache_size <= self._data_count:
            self.record()

    def _record(self):
        """记录数据到文件"""
        if not Path(self.path).exists():
            with open(self.path, 'w'):
                pass

        with open(self.path, 'rb+') as f:
            previous = None
            for i in self._data:
                loc = ByteRecorder.__END if i[1] is None else (i[1], 0)
                if not (previous == loc == ByteRecorder.__END):
                    f.seek(loc[0], loc[1])
                    previous = loc
                f.write(i[0])



# -*- coding:utf-8 -*-
from abc import abstractmethod
from pathlib import Path
from threading import Lock
from time import sleep

from .setter import OriginalSetter, BaseSetter
from .tools import get_usable_path


class OriginalRecorder(object):
    """记录器的基类"""
    SUPPORTS = ('any',)

    def __init__(self, path=None, cache_size=None):
        """
        :param path: 保存的文件路径
        :param cache_size: 每接收多少条记录写入文件，0为不自动写入
        """
        self._data = None
        self._path = None
        self._type = None
        self._lock = Lock()
        self._pause_add = False  # 文件写入时暂停接收输入
        self._pause_write = False  # 标记文件正在被一个线程写入
        self.show_msg = True
        self._setter = None
        self._data_count = 0  # 已缓存数据的条数

        if path:
            self.set.path(path)
        self._cache = cache_size if cache_size is not None else 1000

    def __del__(self):
        """对象关闭时把剩下的数据写入文件"""
        self.record()

    @property
    def set(self):
        """返回用于设置属性的对象"""
        if self._setter is None:
            self._setter = OriginalSetter(self)
        return self._setter

    @property
    def cache_size(self):
        """返回缓存大小"""
        return self._cache

    @property
    def path(self):
        """返回文件路径"""
        return self._path

    @property
    def type(self):
        """返回文件类型"""
        return self._type

    @property
    def data(self):
        """返回当前保存在缓存的数据"""
        return self._data

    def record(self, new_path=None):
        """记录数据，可保存到新文件
        :param new_path: 文件另存为的路径，会保存新文件
        :return: 文件路径
        """
        # 具体功能由_record()实现，本方法实现自动重试及另存文件功能
        original_path = return_path = self._path
        if new_path:
            new_path = str(get_usable_path(new_path))
            return_path = self._path = new_path

            if Path(original_path).exists():
                from shutil import copy
                copy(original_path, self._path)

        if not self._data:
            return return_path

        if not self._path:
            raise ValueError('保存路径为空。')

        with self._lock:
            self._pause_add = True  # 写入文件前暂缓接收数据
            if self.show_msg:
                print(f'{self.path} 开始写入文件，切勿关闭进程。')

            Path(self.path).parent.mkdir(parents=True, exist_ok=True)
            while True:
                try:
                    while self._pause_write:  # 等待其它线程写入结束
                        sleep(.1)

                    self._pause_write = True
                    self._record()
                    break

                except PermissionError:
                    if self.show_msg:
                        print('\r文件被打开，保存失败，请关闭，程序会自动重试。', end='')

                except Exception as e:
                    try:
                        with open('failed_data.txt', 'a+', encoding='utf-8') as f:
                            f.write(str(self.data) + '\n')
                        print('保存失败的数据已保存到failed_data.txt。')
                    except:
                        raise e
                    raise

                finally:
                    self._pause_write = False

                sleep(.3)

            if new_path:
                self._path = original_path

            if self.show_msg:
                print(f'{self.path} 写入文件结束。')
            self.clear()
            self._data_count = 0
            self._pause_add = False

        return return_path

    def clear(self):
        """清空缓存中的数据"""
        if self._data:
            self._data.clear()

    @abstractmethod
    def add_data(self, data):
        pass

    @abstractmethod
    def _record(self):
        pass

    # ---------------即将废弃--------------------
    def set_path(self, path):
        """设置文件路径
        :param path: 文件路径
        :return: None
        """
        self.set.path(path)


class BaseRecorder(OriginalRecorder):
    """Recorder、Filler和DBRecorder的父类"""
    SUPPORTS = ('xlsx', 'csv')

    def __init__(self, path=None, cache_size=None):
        """
        :param path: 保存的文件路径
        :param cache_size: 每接收多少条记录写入文件，0为不自动写入
        """
        super().__init__(path, cache_size)
        self._before = []
        self._after = []
        self._encoding = 'utf-8'
        self._table = None

    @property
    def set(self):
        """返回用于设置属性的对象"""
        if self._setter is None:
            self._setter = BaseSetter(self)
        return self._setter

    @property
    def before(self):
        """返回当前before内容"""
        return self._before

    @property
    def after(self):
        """返回当前after内容"""
        return self._after

    @property
    def table(self):
        """返回默认表名"""
        return self._table

    @property
    def encoding(self):
        """返回编码格式"""
        return self._encoding

    @abstractmethod
    def add_data(self, data, table=None):
        pass

    @abstractmethod
    def _record(self):
        pass

    # ---------------即将废弃--------------------
    def set_before(self, before):
        """设置before
        :param before: before
        :return: None
        """
        self.set.before(before)

    def set_after(self, after):
        """设置after
        :param after: after
        :return: None
        """
        self.set.after(after)



# -*- coding:utf-8 -*-
from pathlib import Path
from typing import Union, List, Any, Tuple, Optional

from .base import BaseRecorder
from .style import CellStyle
from .setter import FillerSetter


class Filler(BaseRecorder):
    _set: FillerSetter = ...
    _key_cols: Optional[List[int], bool] = ...
    _begin_row: Optional[str, int] = ...
    _sign_col: Optional[int, bool] = ...
    _data_col: Optional[int] = ...
    _sign: Any = ...
    _deny_sign: bool = ...
    _link_style: CellStyle = ...
    _quote_char: str = ...
    _delimiter: str = ...
    _data: Union[list, dict] = ...
    data: Union[list, dict] = ...
    row_num_title: str = ...

    def __init__(self, path: Optional[str, Path] = None,
                 cache_size: int = None,
                 key_cols: Union[str, int, list, tuple, bool] = True,
                 begin_row: int = 2,
                 sign_col: Union[str, int, bool] = True,
                 data_col: Optional[int, str] = None,
                 sign: Any = None,
                 deny_sign: bool = False) -> None: ...

    @property
    def sign(self) -> str: ...

    @property
    def deny_sign(self) -> bool: ...

    @property
    def key_cols(self) -> Union[List[int], bool]: ...

    @property
    def sign_col(self) -> Optional[int, bool]: ...

    @property
    def data_col(self) -> int: ...

    @property
    def begin_row(self) -> Union[str, int]: ...

    @property
    def keys(self) -> list: ...

    @property
    def dict_keys(self) -> List[dict]: ...

    @property
    def set(self) -> FillerSetter: ...

    @property
    def delimiter(self) -> str: ...

    @property
    def quote_char(self) -> str: ...

    def add_data(self, data: Any,
                 coord: Union[list, Tuple[Optional[int, str], Union[int, str]], str, int] = 'newline',
                 table: Union[str, bool] = None) -> None: ...

    def set_link(self,
                 coord: Union[int, str, tuple, list],
                 link: Optional[str],
                 content: Optional[int, str, float] = None) -> None: ...

    def set_style(self, coord: Union[int, str, tuple, list], style: Optional[CellStyle],
                  replace: bool = True) -> None: ...

    def set_img(self, coord: Union[int, str, tuple, list], img_path: Optional[str, Path], width: float = None,
                height: float = None) -> None: ...

    def set_row_height(self, row: int, height: float) -> None: ...

    def set_col_width(self, col: Union[int, str], width: float) -> None: ...

    def _record(self) -> None: ...

    def _to_xlsx(self) -> None: ...

    def _to_csv(self) -> None: ...


def get_xlsx_keys(filler: Filler, as_dict: bool) -> List[Union[list, dict]]: ...


def get_csv_keys(filler: Filler, as_dict: bool) -> List[Union[list, dict]]: ...
